/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifdef __GNUG__
#pragma implementation
#endif

#include "linkview.h"
#include "Web.h"

#include <strstream>
#include <algorithm>
#include "copy_line.h"
#include "dfs.h"
#include "NodeMapper.h"

#ifdef STRICT_STL
#define NODEP(x)  (x)
#define LINKP(x)  (x)
#else
#define NODEP(x)  ((node*)(x))
#define LINKP(x)  ((link*)(x))
#endif


Web::Web(
    NodeMapper* mapper):
    graph_(new graph_type),
    active_node_(NULL), active_link_(NULL),
    mapper_(mapper)
{}


Web::~Web()
{
    delete graph_;
}


Web::node*
Web::redirect(
    node* vp_from,
    URL   url)
{
    if (!vp_from) return NULL;
    node v_to(url, vp_from->info());
    node_iterator vi_to = nodes().find(&v_to);
    node* vp_to = NULL;
    if (vi_to != nodes().end()) vp_to = *vi_to;
    else                        vp_to = insert(v_to);
    return redirect(vp_from, vp_to);
}


Web::node*
Web::redirect(
    node* vp_from,
    node* vp_to)
{
    if (!vp_from || !vp_to) return NULL;
    vp_from->info().redirect = vp_to->key();
    vp_from->info().disabled = true;
    while (!vp_from->in().empty()) {
	link_iterator ei = vp_from->in().end();
	--ei;
	link e((*ei)->from(), vp_to, (*ei)->info());
	erase(*ei);
	insert(e);
    }
    while (!vp_from->out().empty()) {
	link_iterator ei = vp_from->out().end();
	--ei;
	link e(vp_to, (*ei)->to(), (*ei)->info());
	erase(*ei);
	insert(e);
    }
    if (active_node() == vp_from) active_node(vp_to);
    return vp_to;
}


void
Web::active_node(
    node* vp)
{
    if (active_node_ == vp) return;
    if (active_node_) {
	active_node_->info().icon_changed  = true;
	active_node_->info().label_changed = true;
    }
    active_node_ = vp;
    vp->info().icon_changed  = true;
    vp->info().label_changed = true;
    active_link_ = NULL;
    update();
}


void
Web::active_link(
    link* ep)
{
    active_link_ = ep;
    update();
}


int
ArrangeNodes::operator()(
    Web::node* vp,
    Web::link*,
    int        depth)
{
    int& level = vp->info().doi;
    int result = (level < 1) ? dfs_visited: dfs_next;
    
    if (vp->info().visited) {
	DEBUGOUT("node " << vp->key() <<": " << "visited");
	return dfs_visited;
    }
    if (dir_ == up   && level > first_level_ - depth || 
	dir_ == down && level < first_level_ - depth) {
	DEBUGOUT("node " << vp->key() <<": " << "already level=" << level);
	vp->info().visited = true;
	return dfs_next;
    }
    vp->info().visited = true;
    if (dir_ == up)     level++;
    else if (level > 0) level--;
    
    web_->mapper()->layout(vp);
    DEBUGOUT("node " << vp->key() << ": "
	     << "lifted up/down to level=" << level);
    return result;
}


void
Web::clear_visited()
{
    for (Web::node_iterator vi = nodes().begin(); vi != nodes().end(); vi++)
	(*vi)->info().visited = false;
}


#if 0
void
Web::layout(
    node* vp)
{
    mapper()->layout(vp);
}
#endif


void
Web::layout_all()
{
    for (node_iterator vi = nodes().begin(); vi != nodes().end(); vi++) {
	mapper()->layout(*vi);
	(*vi)->info().icon_changed = true;
    }
    update();
}


void
Web::lift_up(
    node* vp)
{
    if (!vp) return;
    ArrangeNodes ar_up(this, ArrangeNodes::up, vp->info().doi);
    clear_visited();
    (void) dfs_forward(vp, ar_up, vp->info().doi + 1);
    update();
}


void
Web::lift_down(
    node* vp)
{
    if (!vp) return;
    ArrangeNodes ar_down(this, ArrangeNodes::down, vp->info().doi);
    clear_visited();
    (void) dfs_forward(vp, ar_down);
    update();
}


void
Web::lift_down_all()
{
    for (Web::node_iterator vi = nodes().begin(); vi != nodes().end(); vi++) {
	(*vi)->info().doi = 0;
	mapper()->layout(*vi);
    }
    update();
}


void
Web::set_mark(
    string mark,
    node*  vp)
{
    if (!vp) return;
    mark_map::iterator i = marks_.find(mark);
    if (i != marks_.end()) {
	
	NodeInfo& info = NODEP((*i).second)->info();
	vector<string>::iterator j = 
	    find(info.mark_keys.begin(), info.mark_keys.end(), mark);
	if (j != info.mark_keys.end()) info.mark_keys.erase(j);
    }
    marks_.insert(mark_map::value_type(mark, vp));
    vp->info().mark_keys.push_back(mark);
    vp->info().icon_changed = true;
    update();
}


Web::node*
Web::get_mark(
    string mark)
{
    mark_map::iterator i = marks().find(mark);
    if (i == marks_.end()) return NULL;
    return NODEP((*i).second);
}
