/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifdef __GNUG__
#pragma implementation
#endif

#include "linkview.h"
#include "WebView.h"

#include <math.h>
#include <stack>  /* #include <queue> */
#include <strstream>
#include <stdlib.h>
#include <string.h>
#include "glex.h"
#include "NodeMapper.h"


#define INC_ROT      10.0
#define INC_TRANS     0.1


WebView::WebView(
    WebFrame* frame):
    frame_(frame), parent_(NULL),
    x_(0), y_(0), width_(1), height_(1),
    render_mode_(surface),
    z_board_(- 2.0 * web()->mapper()->def_radius()),
    inc_trans_(INC_TRANS), inc_rot_(INC_ROT),
    level_(0)
{
    rot_    = coord(-80.0,  0.0,  0.0);
    trans_  = coord(  0.0, -0.4, -2.1);
    origin_ = coord(  0.5,  0.5,  0.0);
}


WebView::~WebView()
{
#ifdef PABLO
    close_all_subviews();
#endif
}


void
WebView::update()
{
    update_prefix();
#ifdef PABLO
    for (webview_vector::iterator i = subviews_.begin();
	 i != subviews_.end(); i++) {
	(*i)->update_prefix();
    }
#endif

    draw();
#ifdef PABLO
    for (webview_vector::iterator i = subviews_.begin();
	 i != subviews_.end(); i++) {
	(*i)->draw();
    }
#endif

    update_suffix();
#ifdef PABLO
    for (webview_vector::iterator i = subviews_.begin();
	 i != subviews_.end(); i++) {
	(*i)->update_suffix();
    }
#endif
}


void
WebView::draw()
{
    draw_prefix();
    draw_board();
    draw_nodes();
    draw_cursor();
    // draw_links() must be done at last! because of blending.
    draw_links();
    draw_suffix();
}


void
WebView::draw_nodes()
{
    draw_nodes_prefix();
    for (Web::node_iterator vi = web()->nodes().begin();
	 vi != web()->nodes().end(); vi++) {
	Web::node* vp = *vi;
	int status = node_status(vp);
	if (status & (disabled_bit | masked_bit)) continue;
#if 0
	Web::link_iterator ei; 
	for (ei = vp->in().begin();
	     ei != vp->in().end() &&
	     !((*ei)->from() == web()->active_node() ||
	       web()->active_link() &&
	       (*ei)->from() == web()->active_link()->to());
	     ei++)
	    ;
	if (!(status & (scanned_bit | active_bit | active_link_bit)) &&
	    ei == vp->in().end() && vp->out().size() == 0) {
	    vp->info().drawn = false;
	    continue;
	}
#endif
	vp->info().drawn = true;
	draw_a_node(*vi, status);
    }
    draw_nodes_suffix();
}


void
WebView::draw_links()
{
    draw_links_prefix();
    for (Web::link_iterator ei = web()->links().begin();
	 ei != web()->links().end(); ei++) {
	Web::link* ep = *ei;
	int status = link_status(ep);
	if (status & (disabled_bit | masked_bit)) continue;
#if 0
	if (!(status & (scanned_bit | active_bit | active_link_bit) ||
	      web()->active_link() &&
	      web()->active_link()->to() == ep->from()) &&
	    ep->to()->out().size() == 0) {
	    ep->info().drawn = false;
	    continue;
	}
#endif
	ep->info().drawn = true;
	draw_a_link(ep, status);
    }
    draw_links_suffix();
}


void
WebView::render_mode(
    render_t mode)
{
    render_mode_ = mode;
    for (Web::node_iterator vi = web()->nodes().begin();
	 vi != web()->nodes().end(); vi++)
	(*vi)->info().icon_changed = true;
    update();
}


void
WebView::translate(
    GLfloat& axis,
    GLfloat  inc)
{
    axis += inc;
#ifdef PABLO
    close_all_subviews();
#endif
    update();
}


void 
WebView::rotate(
    GLfloat& axis,
    GLfloat  inc)
{
    axis += inc;
    if      (axis < -180) axis += 360;
    else if (axis >  180) axis -= 360;
#ifdef PABLO
    close_all_subviews();
#endif
    update();
}


void 
WebView::trans_left_right_abs(
    GLfloat dist)
{
    trans_.x = dist;
#ifdef PABLO
    close_all_subviews();
#endif
    update();
}


void 
WebView::trans_up_down_abs(
    GLfloat dist)
{
    trans_.y = dist;
#ifdef PABLO
    close_all_subviews();
#endif
    update();
}


void 
WebView::rot_left_right_abs(
    GLfloat angle)
{
    rot_.z = angle;
#ifdef PABLO
    close_all_subviews();
#endif
    update();
}


void 
WebView::rot_up_down_abs(
    GLfloat angle)
{
    rot_.x = angle;
#ifdef PABLO
    close_all_subviews();
#endif
    update();
}


void
WebView::home_view()
{
    rot_    = coord(-80.0,  0.0,  0.0);
    trans_  = coord(  0.0, -0.4, -2.1);
    origin_ = coord(  0.5,  0.5,  0.0);
#ifdef PABLO
    close_all_subviews();
#endif
    update();
}


void
WebView::top_view()
{
    rot_    = coord(0.0, 0.0,  0.0);
    trans_  = coord(0.0, 0.0, -2.0);
    origin_ = coord(0.5 ,0.5,  0.0);
#ifdef PABLO
    close_all_subviews();
#endif
    update();
}


void
WebView::calc_2d_layout()
{
    for (Web::node_iterator vi = web()->nodes().begin();
	 vi != web()->nodes().end(); vi++) {
	float x, y;
	get_raster_pos((*vi), x, y);
	(*vi)->info().x_2d = x;
	(*vi)->info().y_2d = y;
    }
}


int
WebView::node_status(
    Web::node* vp)
{
    int status = 0;
    NodeInfo& info = vp->info();
    if (info.scanned)               status |= scanned_bit;
    if (web()->active_node() == vp) status |= active_bit;
    if (!info.mark_keys.empty())    status |= marked_bit;
    if (info.disabled)              status |= disabled_bit;
    if (info.r <= 0)                status |= masked_bit;
    if (web()->active_link() && web()->active_link()->to() == vp)
	status |= active_link_bit;
    return status;
}


int
WebView::link_status(
    Web::link* ep)
{
    int status = 0;
    NodeInfo& from = ep->from()->info();
    NodeInfo& to   = ep->to()->info();
    if (to.scanned)                         status |= scanned_bit;
    if (web()->active_node() == ep->from()) status |= active_bit;
    if (from.disabled || to.disabled)       status |= disabled_bit;
    if (from.r <= 0 || to.r <= 0)           status |= masked_bit;
    if (web()->active_link() == ep)         status |= active_link_bit;
    return status;
}
