/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifndef WEBVIEW_H
#define WEBVIEW_H
#ifdef __GNUG__
#pragma interface
#endif

//////////////////////////////////////////////////////////////////////////////
//  webview.h --
//

#include "Web.h"
#include "WebFrame.h"


//////////////////////////////////////////////////////////////////////////////
//  WebView --
//
class WebView
{
 public:
    struct coord
    {
	float x, y, z;
	coord(const coord &a): x(a.x), y(a.y), z(a.z) {}
	coord(float a = 0, float b = 0, float c = 0): x(a), y(b), z(c) {}
    };
    
    // primitive functions
    WebView(WebFrame*);
#ifdef PABLO
    WebView(WebView* view, int x, int y, int width, int height);
#endif
    virtual ~WebView();

    // operations and manipulations to the focused node
    void focus(URL url)  { frame()->focus(url); }
    void lift_up()       { frame()->lift_up(); }
    void lift_down()     { frame()->lift_down(); }
    void lift_down_all() { frame()->lift_down_all(); }
    void browse()        { frame()->browse(); }
    void reload()        { frame()->reload(); }
    void scan()          { frame()->scan(); }

    // mark operations
    void set_mark(string mark, URL url)
	{ web()->set_mark(mark, *web()->nodes().find(&Web::node(url))); }
    void set_mark(string mark)
	{ web()->set_mark(mark, web()->active_node()); }
    void set_mark()
	{ set_mark("system"); }

    // translation operations
    void trans_right   (float d) { translate(trans_.x,  d); }
    void trans_right   ()        { translate(trans_.x,  inc_trans_); }
    void trans_left    (float d) { translate(trans_.x, -d); }
    void trans_left    ()        { translate(trans_.x, -inc_trans_); }
    void trans_up      (float d) { translate(trans_.y,  d); }
    void trans_up      ()        { translate(trans_.y,  inc_trans_); }
    void trans_down    (float d) { translate(trans_.y, -d); }
    void trans_down    ()        { translate(trans_.y, -inc_trans_); }

    // rotation operations
    void rot_down       (float a) { rotate(rot_.x,  a); }
    void rot_down       ()        { rotate(rot_.x,  inc_rot_); }
    void rot_up         (float a) { rotate(rot_.x, -a); }
    void rot_up         ()        { rotate(rot_.x, -inc_rot_); }
    void rot_right      (float a) { rotate(rot_.z,  a); }
    void rot_right      ()        { rotate(rot_.z,  inc_rot_); }
    void rot_left       (float a) { rotate(rot_.z, -a); }
    void rot_left       ()        { rotate(rot_.z, -inc_rot_); }

    void zoom_in (float d) { translate(trans_.z,  d); }
    void zoom_in ()        { translate(trans_.z,  inc_trans_); }
    void zoom_out(float d) { translate(trans_.z, -d); }
    void zoom_out()        { translate(trans_.z, -inc_trans_); }

    void trans_left_right_abs(float);
    void trans_up_down_abs(float);
    void rot_left_right_abs(float);
    void rot_up_down_abs(float);
    
    
    // rendering mode
    enum render_t { wireframe, surface, face_icon };
    render_t render_mode() const   { return render_mode_; }
    void     render_mode(render_t m);
    void     render_wireframe()    { render_mode(wireframe); }
    void     render_surface()      { render_mode(surface); }
    void     render_face_icon()    { render_mode(face_icon); }

    // view control
    void home_view();
    void top_view();
    void focus_view() {}

    // picking objects
    virtual void pick(int x, int y, vector<int>& sysobjs,
		      Web::node_vector& nodes, Web::link_vector& links) = 0;
    virtual void viewable_nodes(Web::node_vector&) = 0;

    // update 
    void update();
    void message(string s) { frame()->message(s); }

    // functions for debug and evaluation
    void calc_2d_layout();
    virtual void get_raster_pos(Web::node* v, float& x, float& y) = 0;

#ifdef PABLO
    WebView* open_subview(int x, int y, int width, int height);
    void     close_subview(WebView* view);
    void     close_all_subviews();
    void     close();
#endif

    // member access
    WebFrame* frame()    { return frame_; }
    WebView*  rootview() { return frame_->rootview(); }
    Web*      web()      { return frame()->web(); }
    WebView*  parent()   { return parent_; }

    int   xpos()      const { return x_; }
    int   ypos()      const { return y_; }
    int   width()     const { return width_; }
    int   height()    const { return height_; }
    float z_board()   const { return z_board_; }
    coord trans()     const { return trans_; }
    coord rot()       const { return rot_; }
    coord origin()    const { return origin_; }
    float inc_trans() const { return inc_trans_; }
    float inc_rot()   const { return inc_rot_; }
    int   level()     const { return level_; }

 protected:
    enum { scanned_bit  = 001,  active_bit = 002,  marked_bit = 004,
	   disabled_bit = 010,  masked_bit = 020,  active_link_bit = 040, };
    int node_status(Web::node*);
    int link_status(Web::link*);

    // subroutines of update()
    virtual void update_prefix() {}
    virtual void update_suffix() {}
    void         draw();

    // subroutines of draw()
    virtual void draw_prefix() {}
    virtual void draw_suffix() {}
    virtual void draw_board() {}
    virtual void draw_cursor() {}

    void         draw_nodes();
    virtual void draw_nodes_prefix() {}
    virtual void draw_nodes_suffix() {}
    virtual void draw_a_node(Web::node* vp, int status) = 0;
    void         draw_a_node_aux(void* p, int s)
	{ draw_a_node((Web::node*) p, s); }

    void         draw_links();
    virtual void draw_links_prefix() {}
    virtual void draw_links_suffix() {}
    virtual void draw_a_link(Web::link* ep, int status) = 0;
    void         draw_a_link_aux(void* p, int s)
	{ draw_a_link((Web::link*) p, s); }

    // subroutines of trans_*() and rot_*()
    void translate(float& axis, float inc);
    void rotate(float& axis, float inc);
#ifdef PABLO
    virtual WebView* do_open_subview(int x, int y, int width, int height) = 0;
    void    set_origin();
#endif

    virtual void resize(int x, int y, int w, int h)
	{ x_ = x;  y_ = y;  width_ = w;  height_ = h; }

 private:
    WebFrame* frame_;
    WebView*  parent_;
    int       x_;
    int       y_;
    int       width_;
    int       height_;
    render_t  render_mode_;
    float     z_board_;
    coord     trans_;
    coord     rot_;
    coord     origin_;
    float     inc_trans_;
    float     inc_rot_;
    int       level_;

    typedef ptrvector<WebView*> webview_vector; 
    webview_vector subviews_;
};

#endif /* WEBVIEW_H */
