/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1996  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                      Keio University (Matsushita Lab.)
//                      All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifndef GRAPH_H
#define GRAPH_H
#ifdef __GNUG__
//#pragma interface
#endif

#include <set>
#include <map>
#include <algorithm>
#include "ptrvector.h"


#ifdef STRICT_STL

template <class VP>
struct vertex_less: binary_function<VP, VP, bool>
{
    bool operator()(VP v1, VP v2) const { return v1->key() < v2->key(); }
};

template <class VP>
struct vertex_equal_to: binary_function<VP, VP, bool>
{
    bool operator()(VP v1, VP v2) const { return v1->key() == v2->key(); }
};

#else

template <class VP>
struct vertex_less: binary_function<void*, void*, bool>
{
    bool operator()(void* v1, void* v2) const
	{ return VP(v1)->key() < VP(v2)->key(); }
};

template <class VP>
struct vertex_equal_to: binary_function<void*, void*, bool>
{
    bool operator()(void* v1, void* v2) const
	{ return VP(v1)->key() == VP(v2)->key(); }
};

#endif /* STRICT_STL */


//////////////////////////////////////////////////////////////////////////////
//  graph -- template class to represent graph structures
//
template <class VKey, class Vertex, class Edge>
class graph
{
    typedef graph<VKey, Vertex, Edge> self_type;
 public:
    // types
    class vertex;
    class edge;
    typedef ptrset<vertex*, vertex_less<vertex*> > vertex_container;
    typedef ptrvector<edge*> edge_container;
    typedef vertex_container::iterator vertex_iterator;
    typedef edge_container::iterator   edge_iterator;
    typedef vertex_container::reverse_iterator vertex_reverse_iterator;
    typedef edge_container::reverse_iterator   edge_reverse_iterator;
    
    //////////////////////////////////////////////////////////////////////////
    //  vertex
    //
    class vertex
    {
	friend class graph;
     public:
	// premitive functions
	vertex(const VKey key): key_(key) {}
	vertex(const VKey key, const Vertex& info): key_(key), info_(info) {}

	// member access
	const VKey&     key()  { return this->key_; }
	edge_container& in()   { return this->in_; }
	edge_container& out()  { return this->out_; }
	Vertex&         info() { return this->info_; }

     private:
	VKey           key_;  // id
	edge_container in_;   // incomming edges
	edge_container out_;  // outgoing edges
	Vertex         info_; // additional infomation
    };

    //////////////////////////////////////////////////////////////////////////
    //  edge
    //
    class edge
    {
	friend class graph;
     public:
	// premitive functions
	edge(vertex* from, vertex* to): from_(from), to_(to) {}
	edge(vertex* from, vertex* to, const Edge& info):
	    from_(from), to_(to), info_(info) {}

	// member access
	vertex* from() { return this->from_; }
	vertex* to()   { return this->to_; }
	Edge&   info() { return this->info_; }

     private:
	vertex* from_;  // source vertex
	vertex* to_;    // destination vertex
	Edge    info_;  // additional infomation
    };
    
 public:
    vertex_container& verteces() { return verteces_; }
    edge_container&   edges()    { return edges_; }

    //
    vertex* insert(const vertex& v)
    {
	vertex* vp = new vertex(v);
	vertex_iterator i = verteces_.find(vp);
	if (i != verteces_.end()) verteces_.erase(i);
	verteces_.insert(vp);
	return vp;
    }

    edge* insert(const edge& e)
    {
	edge* ep = new edge(e);
	edges_.push_back(ep);
	e.from_->out_.push_back(ep);
	e.to_->in_.push_back(ep);
	return ep;
    }

    void erase(vertex* vp)
    {
	vertex_iterator i = verteces_.find(vp);
	if (i != verteces_.end()) verteces_.erase(i);
	while (!vp->in_.empty())  erase(vp->in_.back());
	while (!vp->out_.empty()) erase(vp->out_.back());
	delete vp;
    }

    void erase(edge* ep)
    {
	edge_iterator i;
	i = find(edges_.begin(), edges_.end(), ep);
	if (i != edges_.end()) edges_.erase(i);
	i = find(ep->from_->out_.begin(), ep->from_->out_.end(), ep);
	if (i != ep->from_->out_.end()) ep->from_->out_.erase(i);
	i = find(ep->to_->in_.begin(), ep->to_->in_.end(), ep);
	if (i != ep->to_->in_.end()) ep->to_->in_.erase(i);
	delete ep;
    }

 private:
    vertex_container verteces_;
    edge_container   edges_;
};

#endif /* GRAPH_H */
