/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1996-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifdef __GNUG__
#pragma implementation
#endif

#include "linkview.h"
#include "www.h"
#include <string.h>
#include <strstream>
#include "kanji.h"
#include "HTMLexDTD.h"

#ifdef JAPANESE
#define TOEUC anytoeuc
#else
#define TOEUC
#endif


URL::URL(
    HTAnchor* htanchor):
    absolute_(NULL)
{
    char* str = HTAnchor_address(htanchor);
    HTSimplify(&str);
    absolute_ = str;
    HT_FREE(str);
}


string
URL::htparse(
    string url,
    string base)
{
    if (base.length() == 0) return htparse(url);
    char* str = HTParse(url.c_str(), base.c_str(), PARSE_ALL);
    HTSimplify(&str);
    string ret = str;
    HT_FREE(str);
    return ret;
}


string
URL::htparse(
    string url)
{
    char* str = HTParse(url.c_str(), NULL, PARSE_ALL);
    HTSimplify(&str);
    string ret = str;
    HT_FREE(str);
    return ret;
}


string
URL::parse(
    int fields)
{
    char* str = HTParse(absolute_.c_str(), NULL, fields);
    string ret = str;
    HT_FREE(str);
    return ret;
}


string
URL::relative(
    string base) const
{
    char* str = HTRelative(absolute_.c_str(), base.c_str());
    string ret = str;
    HT_FREE(str);
    return ret;
}



string
URL::last_word() const
{
    string str = absolute_;

    size_t pos = str.rfind("/index.");
    if (pos == NPOS) pos = str.rfind("/welcome.");
    if (pos == NPOS) pos = str.rfind("/Welcome.");
    if (pos != NPOS && pos == str.rfind('/')) {
	if (pos > 1) str.remove(pos);
	else         str = '/';
    }

    while (str.length() > 1 && str[str.length() - 1] == '/')
	str.remove(str.length() - 1);
    
    pos = str.rfind('/');
    if (pos > 1) str.remove(0, pos + 1);

#if 0
    if (str[0] == '~') str.remove(0, 1);

    if (str.compare("www.", 4) == 0 || str.compare("web.", 4) == 0)
	str.remove(0, 4);
#endif
    return str;
}


SGMLParser::SGMLParser(
    string          name,
    const SGML_dtd& dtd,
    HTRequest*      req):
    class_name_(name)
{
    structured_class_.name          = class_name_.c_str();
    structured_class_.flush         = ht_flush;
    structured_class_._free         = ht_free;
    structured_class_.abort         = ht_abort;
    structured_class_.put_character = ht_put_character;
    structured_class_.put_string    = ht_put_string;
    structured_class_.put_block     = ht_put_block;
    structured_class_.start_element = ht_start_element;
    structured_class_.end_element   = ht_end_element;
    structured_class_.put_entity    = ht_put_entity;

    me_ = new HTStructured;
    me_->isa         = &structured_class_;
    me_->wrapper     = this;
    htstream_ = SGML_new(&dtd, me_);
    htrequest_ = req;
}


SGMLParser::~SGMLParser()
{
    delete me_;
    HT_FREE(htstream_);
}


int
SGMLParser::free()
{
#if 0
    if (me_->target) {
	int status = (*me_->target->isa->_free)(me_->target);
	if (status != HT_WOULD_BLOCK) me_->target = NULL;
	return status;
    }
#endif
    return HT_OK;
}


int
SGMLParser::abort(HTList*)
{
#if 0
    if (me_->target) {
	(*me_->target->isa->abort)(me_->target, NULL);
	me_->target = NULL;
    }
#endif
    DEBUGOUT("SGMLParser aborting...");
    return HT_ERROR;
}


URL
SGMLParser::current_url() const
{
    return (string) HTAnchor_address((HTAnchor*) HTRequest_anchor(htrequest_));
}


HTMLParser::HTMLParser(
    HTRequest* req):
    SGMLParser("HTMLParser", HTMLex_dtd, req),
    in_header_(false), isindex_(false), in_heading_(H_none), in_anchor_(false),
    prev_char_(' ')
{
    reset();
}


string
HTMLParser::content()
{
    content_ << ends;
    strstream buf;
    anytoeuc(content_, buf);
    string s = buf.str();
    buf.freeze(false);
    return s;
}


void
HTMLParser::reset()
{
    in_header_  = true;
    in_heading_ = H_none;
    in_anchor_  = false;
    prev_char_  = ' ';
    h_str_ = anchor_ = href_url_ = href_title_ = href_rel_ = "";
    base_ = "";
}


URL
HTMLParser::url(
    string relative)
{
    return URL(relative, base_ ? base_: current_url());
}


int
HTMLParser::put_character(
    char c)
{
    if (isascii((unsigned char) c) && isspace(c)) {
	if (isspace(prev_char_ )) {
	    prev_char_ = ' ';
	    return HT_OK;
	}
	c = ' ';
    }
    if (in_heading_) h_str_  += c;
    if (in_anchor_)  anchor_ += c;
    prev_char_ = c;
    content_ << c;
    return HT_OK;
}


void
HTMLParser::start_element(
    int	         element_number,
    const char*  present,
    const char** value)
{
    string alt;
    
    switch (element_number) {
     case HTMLex_TITLE: in_heading_ = TITLE; h_str_ = "";  break;
     case HTMLex_H1:    in_heading_ = H1;    h_str_ = "";  break;
     case HTMLex_H2:    in_heading_ = H2;    h_str_ = "";  break;
     case HTMLex_H3:    in_heading_ = H3;    h_str_ = "";  break;
     case HTMLex_H4:    in_heading_ = H4;    h_str_ = "";  break;
     case HTMLex_H5:    in_heading_ = H5;    h_str_ = "";  break;
     case HTMLex_H6:    in_heading_ = H6;    h_str_ = "";  break;
     case HTMLex_H7:    in_heading_ = H7;    h_str_ = "";  break;

     case HTMLex_A:
	if (present[HTMLex_A_HREF]) {
	    in_anchor_ = true;
	    href_url_ = value[HTMLex_A_HREF];
	    if (present[HTMLex_A_TITLE]) href_title_ = value[HTMLex_A_TITLE];
	    if (present[HTMLex_A_REL])   href_rel_   = value[HTMLex_A_REL];
	}
	break;

     case HTMLex_BASE:
	base_ = URL(value[HTMLex_BASE_HREF], current_url());
	break;

     case HTMLex_IMG:
	alt = "";
	if (present[HTMLex_IMG_ALT]) alt = TOEUC(value[HTMLex_IMG_ALT]);

	if (present[HTMLex_IMG_DYNSRC])
	    movie(url(value[HTMLex_IMG_SRC]), alt);
	else if (present[HTMLex_IMG_SRC])
	    image(url(value[HTMLex_IMG_SRC]), alt);

	if (in_anchor_) {
	    if (present[HTMLex_IMG_ALT]) {
		put_string(value[HTMLex_IMG_ALT]);
	    }
	    else {
		if (present[HTMLex_IMG_DYNSRC]) {
		    strstream buf;
		    buf << '[' << value[HTMLex_IMG_DYNSRC] << ']' << ends;
		    put_string(buf.str());
		    buf.freeze(false);
		}
		else if (present[HTMLex_IMG_SRC]) {
		    strstream buf;
		    buf << "[" << value[HTMLex_IMG_SRC] << ']' << ends;
		    put_string(buf.str());
		    buf.freeze(false);
		}
	    }
	}
	break;

     case HTMLex_FIG:
	if (present[HTMLex_FIG_SRC]) image(url(value[HTMLex_FIG_SRC]));
	break;

     case HTMLex_IMAGE:
	if (present[HTMLex_IMAGE_SRC]) image(url(value[HTMLex_IMAGE_SRC]));
	break;

     case HTMLex_FRAME:
	if (present[HTMLex_FRAME_SRC]) frame(url(value[HTMLex_FRAME_SRC]));
	break;

     case HTMLex_AREA:
	if (present[HTMLex_AREA_HREF]) frame(url(value[HTMLex_AREA_HREF]));
	break;

     case HTMLex_ISINDEX:
   	isindex_ = true;
	break;
    }
}


string
HTMLParser::trim_space(
    string str)
{
    str = TOEUC(str);
    while (str.length() > 0 && str[0] == ' ')
	str.remove(0, 1);
    while (str.length() > 0 && str[str.length() - 1] == ' ')
	str.remove(str.length() - 1, 1);
    return str;
}




void
HTMLParser::end_element(
    int element_number)
{
    switch (element_number) {
     case HTMLex_TITLE:  title(trim_space(h_str_));  in_heading_=H_none; break;
     case HTMLex_H1: heading(H1,trim_space(h_str_)); in_heading_=H_none; break;
     case HTMLex_H2: heading(H2,trim_space(h_str_)); in_heading_=H_none; break;
     case HTMLex_H3: heading(H3,trim_space(h_str_)); in_heading_=H_none; break;
     case HTMLex_H4: heading(H4,trim_space(h_str_)); in_heading_=H_none; break;
     case HTMLex_H5: heading(H5,trim_space(h_str_)); in_heading_=H_none; break;
     case HTMLex_H6: heading(H6,trim_space(h_str_)); in_heading_=H_none; break;
     case HTMLex_H7: heading(H7,trim_space(h_str_)); in_heading_=H_none; break;

     case HTMLex_A:
	if (in_anchor_) {
	    hyperlink(URL(href_url_, current_url()),
		      trim_space(anchor_), trim_space(href_title_), href_rel_);
	}
	href_url_ = anchor_ = href_title_ = href_rel_ = "";
	in_anchor_ = false;
	break;
    }
}
